import netgen.gui
from ngsolve import *
from netgen.csg import *
from math import pi
from netgen.geom2d import unit_square

mesh = Mesh(unit_square.GenerateMesh(maxh=0.5))

SetNumThreads(4)

#exact u
u1=sin(pi*x)*sin(pi*y)-4/pi**2
u2=cos(pi*x)*cos(pi*y)
uexact = CoefficientFunction((u1, u2))
#exact p
pexact = sin(pi*x)*cos(pi*y)

graduexact = CoefficientFunction((u1.Diff(x), u1.Diff(y), \
                                  u2.Diff(x), u2.Diff(y)), dims=(2,2))
gradpex = CoefficientFunction( (pexact.Diff(x), pexact.Diff(y)) )

#source f
source = CoefficientFunction( (-(graduexact[0,0].Diff(x) + graduexact[0,1].Diff(y)) + gradpex[0], -(graduexact[1,0].Diff(x) + graduexact[1,1].Diff(y)) + gradpex[1]) )
#f1=2*(pi**2)*sin(pi*x)*sin(pi*y)+pi*cos(pi*x)*cos(pi*y)
#f2=2*(pi**2)*cos(pi*x)*cos(pi*y)-pi*sin(pi*x)*sin(pi*y)
#source=CoefficientFunction((f1,f2))

k = 2
V = VectorH1(mesh,order=k)
Q = H1(mesh,order=k-1)
R = NumberSpace(mesh) 
X = FESpace([V, Q, R, R])

gfu = GridFunction(X)

(u,p, lam1, lam2), (v,q, mu1, mu2) = X.TnT()
lam = CoefficientFunction((lam1, lam2))
mu = CoefficientFunction((mu1, mu2))

n = specialcf.normal(mesh.dim)


I = Id(2)
g = (-graduexact + pexact * I)

a = BilinearForm(X, symmetric=True)
a += (InnerProduct(grad(u),grad(v))-div(v)*p-div(u)*q)*dx
a += (lam*v+mu*u) * dx 

f = LinearForm(X)
f += source*v*dx
f += - InnerProduct(g*n,v)*ds(definedon=mesh.Boundaries("top|bottom|left|right"))

res = f.vec.CreateVector()

with TaskManager():
    a.Assemble()
    f.Assemble()
    inv = a.mat.Inverse(freedofs=X.FreeDofs(), inverse="umfpack")

    res.data = f.vec - a.mat*gfu.vec
    gfu.vec.data += inv * res

#Solve Stokes function
gfu = GridFunction(X)

output = []
def SolveStokes():
    X.Update()
    gfu.Update()

    with TaskManager():
        #Linear forms
        a.Assemble()
        f.Assemble()
        
        # Solve
        inv = a.mat.Inverse(freedofs=X.FreeDofs(), inverse="umfpack")
        res = f.vec.CreateVector()
        res.data = f.vec - a.mat*gfu.vec
        gfu.vec.data += inv * res

    velocity = gfu.components[0]
    Draw(velocity,mesh,"u")
    pressure = gfu.components[1]
    Draw(pressure,mesh,"p")

    #error
    u,p,lam1,lam2=gfu.components
    err_u =  sqrt (Integrate ((u-uexact)**2, mesh))
    err_p =  sqrt (Integrate ((p-pexact)**2, mesh))
    output.append ((X.ndof, err_u,err_p))
    print(output)

while X.ndof < 15000:
    mesh.Refine()
    SolveStokes()
#Make a table of convergence
i = 1
while i < len(output):
    order_u = log(output[i-1][1]/output[i][1])/log(2)
    order_p = log(output[i-1][2]/output[i][2])/log(2)
    #print order_u, error_u, order_p, error_p
    print("%0.3f | %0.3e || %0.3f | %0.3e " %
      (order_u, output[i-1][1], order_p, output[i-1][2]))
    i =  i+1
